signature FMLA_DT = sig
  include FMLA

  (* specialization datatypes *)
  datatype ('a,'b,'c,'d,'e) DFmla = VarFmla of string | NotFmla of 'a
                                  | TrueFmla | AndFmla of 'b * 'c
                                  | FalseFmla | OrFmla of 'd * 'e
  datatype DAtom = VarAtom of string
  datatype 'a DLit = VarLit of string | NotLit of 'a
  datatype ('b, 'c) DConj = TrueConj | AndConj of 'b * 'c
  datatype ('d, 'e) DDnf = FalseDnf | OrDnf of 'd * 'e

  (* injections *)
  val injFmla : ('a AFmla, 'b AFmla, 'c AFmla, 'd AFmla, 'e AFmla) DFmla -> CFmla
  val injLit : 'a AAtom DLit -> CLit
  val injAtom : DAtom -> CAtom
  val injConj : ('a ALit, 'b AConj) DConj -> CConj
  val injDnf : ('a AConj, 'b ADnf) DDnf -> CDnf

  (* projections *)
  val prjFmla : 'a AFmla -> (CFmla, CFmla, CFmla, CFmla, CFmla) DFmla
  val prjLit : 'a ALit -> CAtom DLit
  val prjAtom : 'a AAtom -> DAtom
  val prjConj : 'a AConj -> (CLit, CConj) DConj
  val prjDnf : 'a ADnf -> (CConj, CDnf) DDnf

  (* maps *)
  val mapFmla : (('a1 -> 'a2) * 
                 ('b1 -> 'b2) * ('c1 -> 'c2) * 
                 ('d1 -> 'd2) * ('e1 -> 'e2)) ->
                ('a1,'b1,'c1,'d1,'e1) DFmla -> ('a2,'b2,'c2,'d2,'e2) DFmla
  val mapAtom : DAtom -> DAtom
  val mapLit : ('a1 -> 'a2) -> 'a1 DLit -> 'a2 DLit
  val mapConj : (('b1 -> 'b2) * ('c1 -> 'c2)) -> ('b1,'c1) DConj -> ('b2,'c2) DConj
  val mapDnf : (('d1 -> 'd2) * ('e1 -> 'e2)) -> ('d1,'e1) DDnf -> ('d2,'e2) DDnf
end
