structure FmlaDT : FMLA_DT = struct
  open Fmla

  (* specialization datatypes *)
  datatype ('a,'b,'c,'d,'e) DFmla = VarFmla of string | NotFmla of 'a
                                  | TrueFmla | AndFmla of 'b * 'c
                                  | FalseFmla | OrFmla of 'd * 'e
  datatype DAtom = VarAtom of string
  datatype 'a DLit = VarLit of string | NotLit of 'a
  datatype ('a, 'b) DConj = TrueConj | AndConj of 'a * 'b
  datatype ('a, 'b) DDnf = FalseDnf | OrDnf of 'a * 'b

  (* injections *)
  fun injFmla f =
    case f of VarFmla s => varFmla s | NotFmla f => notFmla f
            | TrueFmla => trueFmla | AndFmla (f1, f2) => andFmla (f1, f2)
            | FalseFmla => falseFmla | OrFmla (f1, f2) => orFmla (f1, f2)
  fun injLit f = case f of VarLit s => varLit s | NotLit f => notLit f
  fun injAtom f = case f of VarAtom s => varAtom s
  fun injConj f = case f of TrueConj => trueConj | AndConj (f1, f2) => andConj (f1, f2)
  fun injDnf f = case f of FalseDnf => falseDnf | OrDnf (f1, f2) => orDnf (f1, f2)

  (* projections *)
  fun prjFmla f = caseFmla f {Var=VarFmla, Not=NotFmla,
                              True=fn () => TrueFmla, And=AndFmla,
                              False=fn () => FalseFmla, Or=OrFmla}
  fun prjLit f = caseLit f {Var=VarLit, Not=NotLit}
  fun prjAtom f = caseAtom f {Var=VarAtom}
  fun prjConj f = caseConj f {True=fn () => TrueConj, And=AndConj}
  fun prjDnf f = caseDnf f {False=fn () => FalseDnf, Or=OrDnf}

  (* Maps *)
  fun mapFmla (F1,F2,F3,F4,F5) f = 
    case f of VarFmla s => VarFmla s | NotFmla f => NotFmla (F1 f)
            | TrueFmla => TrueFmla | AndFmla (f1, f2) => AndFmla (F2 f1, F3 f2)
            | FalseFmla => FalseFmla | OrFmla (f1, f2) => OrFmla (F4 f1, F5 f2)
  fun mapAtom f = case f of VarAtom s => VarAtom s
  fun mapLit F f = case f of VarLit s => VarLit s | NotLit f => NotLit (F f)
  fun mapConj (F1,F2) f = 
    case f of TrueConj => TrueConj | AndConj (f1, f2) => AndConj (F1 f1, F2 f2)
  fun mapDnf (F1,F2) f = 
    case f of FalseDnf => FalseDnf | OrDnf (f1, f2) => OrDnf (F1 f1, F2 f2)
end