signature FMLA = sig
  (* specialization type *)
  type 'a t

  (* abstract types *)
  type 'a AFmla = {fmla: 'a} t
   and 'a ALit  = {fmla: {lit: 'a}} t           
   and 'a AAtom = {fmla: {lit: {atom: 'a}}} t
   and 'a AConj = {fmla: {conj: 'a}} t          
   and 'a ADnf  = {fmla: {dnf: 'a}} t

  (* concrete types *)
  type CFmla = unit AFmla
   and CLit  = unit ALit            and CAtom = unit AAtom
   and CConj = unit AConj           and CDnf = unit ADnf

  (* constructors *)
  val varFmla   : string -> CFmla   val notFmla : 'a AFmla -> CFmla
  val trueFmla  : CFmla             val andFmla : 'a AFmla * 'b AFmla -> CFmla
  val falseFmla : CFmla             val orFmla  : 'a AFmla * 'b AFmla  -> CFmla
  val varLit    : string -> CLit    val notLit  : 'a AAtom -> CLit
  val varAtom   : string -> CAtom
  val trueConj  : CConj             val andConj : 'a ALit * 'b AConj -> CConj
  val falseDnf  : CDnf              val orDnf   : 'a AConj * 'b ADnf  -> CDnf

  (* destructors *)
  val caseFmla : 'a AFmla -> {Var:string -> 'b, Not:CFmla -> 'b,
                              True:unit -> 'b,  And:CFmla * CFmla -> 'b,
                              False:unit -> 'b, Or:CFmla * CFmla -> 'b} -> 'b
  val caseLit  : 'a ALit -> {Var:string -> 'b, Not:CAtom -> 'b} -> 'b
  val caseAtom : 'a AAtom -> {Var:string -> 'b} -> 'b
  val caseConj : 'a AConj -> {True:unit -> 'b, And:CLit * CConj -> 'b} -> 'b
  val caseDnf  : 'a ADnf -> {False:unit -> 'b, Or:CConj * CDnf -> 'b} -> 'b

  (* coercions *)
  val coerceFmla : 'a AFmla -> CFmla
  val coerceLit  : 'a ALit -> CLit   
  val coerceAtom : 'a AAtom -> CAtom
  val coerceConj : 'a AConj -> CConj
  val coerceDnf  : 'a ADnf -> CDnf
end