structure Fmla :> FMLA = struct
  (* representation type *)
  datatype t' = Var of string | Not of t'
              | True | And of t' * t'
              | False | Or of t' * t'
  (* specialization type *)
  type 'a t = t'

  (* abstract types *)
  type 'a AFmla = {fmla: 'a} t
   and 'a ALit = {fmla: {lit: 'a}} t
   and 'a AAtom = {fmla: {lit: {atom: 'a}}} t
   and 'a AConj = {fmla: {conj: 'a}} t
   and 'a ADnf = {fmla: {dnf: 'a}} t

  (* concrete types *)
  type CFmla = unit AFmla
   and CLit = unit ALit             and CAtom = unit AAtom
   and CConj = unit AConj           and CDnf = unit ADnf

  (* constructors *)
  val varFmla = Var                 val notFmla = Not
  val trueFmla = True               val andFmla = And
  val falseFmla = False             val orFmla = Or
  val varLit = Var                  val notLit = Not
  val varAtom = Var
  val trueConj = True               val andConj = And
  val falseDnf = False              val orDnf = Or

  (* destructors *)
  fun fail _ = raise Match
  fun casef c (v,n,t,a,f,o) = case c of Var(s) => v (s) | Not(f) => n (f) 
                                      | True => t () | And(f,g) => a (f,g) 
                                      | False => f () | Or(f,g) => o (f,g)
  fun caseFmla c {Var=v,Not=n,True=t,And=a,False=f,Or=o} = casef c (v,n,t,a,f,o)
  fun caseLit c {Var=v,Not=n} = casef c (v,n,fail,fail,fail,fail)
  fun caseAtom c {Var=v} = casef c (v,fail,fail,fail,fail,fail)
  fun caseConj c {True=t,And=a} = casef c (fail,fail,t,a,fail,fail)
  fun caseDnf c {False=f,Or=o} = casef c (fail,fail,fail,fail,f,o)

  (* coercions *)
  fun coercef (v) = v
  val coerceFmla = coercef
  val coerceLit = coercef           val coerceAtom = coercef
  val coerceConj = coercef          val coerceDnf = coercef
end